package jobshop.solvers;

import jobshop.Instance;
import jobshop.Result;
import jobshop.encodings.ResourceOrder;
import jobshop.encodings.Schedule;
import jobshop.encodings.Task;

import java.util.ArrayList;
import java.util.Iterator;

import static jobshop.Result.ExitCause.ProvedOptimal;

/** An empty shell to implement a greedy solver. */
public class GreedySolver implements Solver {

    /** All possible priorities for the greedy solver. */
    public enum Priority {
        SPT, LPT, SRPT, LRPT, EST_SPT, EST_LPT, EST_SRPT, EST_LRPT
    }

    /** Priority that the solver should use. */
    final Priority priority;

    final ArrayList<Task> tachesRestantes;
    int[] machineFreeDate;
    int[] jobFreeDate;

    int pts = 10;


    /** Creates a new greedy solver that will use the given priority. */
    public GreedySolver(Priority p) {
        this.priority = p;
        this.tachesRestantes = new ArrayList<>();
    }

    private int getEarliestDate(Instance instance, Task task) {
        return Math.max(machineFreeDate[instance.machine(task)], jobFreeDate[task.job]);
    }

    /** return true if t1 est plus prioritaire que t1 **/
    private boolean prioritaire(Instance instance, Task t1, Task t2) {
        boolean rt = false;
        switch(priority) {
            case SPT:
                rt = instance.duration(t1) <= instance.duration(t2);
                break;
            case LPT:
                rt = instance.duration(t1) >= instance.duration(t2);
                break;
            case SRPT:
                int i;
                int sigmaT1 = 0;
                int sigmaT2 = 0;
                for (i=t1.task; i<instance.numTasks; i++) {
                    sigmaT1 += instance.duration(t1.job, i);
                }
                for (i=t2.task; i<instance.numTasks; i++) {
                    sigmaT2 += instance.duration(t2.job, i);
                }
                rt = sigmaT1 <= sigmaT2;
                break;
            case LRPT:
                sigmaT1 = 0;
                sigmaT2 = 0;
                for (i=t1.task; i<instance.numTasks; i++) {
                    sigmaT1 += instance.duration(t1.job, i);
                }
                for (i=t2.task; i<instance.numTasks; i++) {
                    sigmaT2 += instance.duration(t2.job, i);
                }
                rt = sigmaT1 >= sigmaT2;
                break;
            case EST_SPT:
                if (getEarliestDate(instance, t1) < getEarliestDate(instance, t2)) {
                    rt = true;
                } else if (getEarliestDate(instance, t1) > getEarliestDate(instance, t2)) {
                    rt = false;
                } else {
                    rt = instance.duration(t1) <= instance.duration(t2);
                }
                break;
            case EST_LPT:
                if (getEarliestDate(instance, t1) < getEarliestDate(instance, t2)) {
                    rt = true;
                } else if (getEarliestDate(instance, t1) > getEarliestDate(instance, t2)) {
                    rt = false;
                } else {
                    rt = instance.duration(t1) >= instance.duration(t2);
                }
                break;
            case EST_SRPT:
                if (getEarliestDate(instance, t1) < getEarliestDate(instance, t2)) {
                    rt = true;
                } else if (getEarliestDate(instance, t1) > getEarliestDate(instance, t2)) {
                    rt = false;
                } else {
                    sigmaT1 = 0;
                    sigmaT2 = 0;
                    for (i=t1.task; i<instance.numTasks; i++) {
                        sigmaT1 += instance.duration(t1.job, i);
                    }
                    for (i=t2.task; i<instance.numTasks; i++) {
                        sigmaT2 += instance.duration(t2.job, i);
                    }
                    rt = sigmaT1 <= sigmaT2;
                }
                break;
            case EST_LRPT:
                if (getEarliestDate(instance, t1) < getEarliestDate(instance, t2)) {
                    rt = true;
                } else if (getEarliestDate(instance, t1) > getEarliestDate(instance, t2)) {
                    rt = false;
                } else {
                    sigmaT1 = 0;
                    sigmaT2 = 0;
                    for (i=t1.task; i<instance.numTasks; i++) {
                        sigmaT1 += instance.duration(t1.job, i);
                    }
                    for (i=t2.task; i<instance.numTasks; i++) {
                        sigmaT2 += instance.duration(t2.job, i);
                    }
                    rt = sigmaT1 >= sigmaT2;
                }
                break;
        }
        return rt;
    }

    private void addTask(Instance instance, Task task) {
        if (priority == Priority.SPT || priority == Priority.LPT || priority == Priority.SRPT || priority == Priority.LRPT) {
            Iterator<Task> iter = this.tachesRestantes.iterator();
            int index = 0;
            boolean trouve = false;
            while (iter.hasNext() && !trouve) {
                Task current = iter.next();
                if (this.prioritaire(instance, task, current)) {
                    trouve = true;
                } else {
                    index++;
                }
            }
            this.tachesRestantes.add(index, task);
        } else {
            this.tachesRestantes.add(task);
        }
    }

    private Task getTask(Instance instance) {
        if (priority == Priority.SPT || priority == Priority.LPT || priority == Priority.SRPT || priority == Priority.LRPT) {
            return this.tachesRestantes.remove(0);
        } else {
            int indexBest = 0;
            int indexChallenger;
            for (indexChallenger = 1; indexChallenger < tachesRestantes.size(); indexChallenger++) {
                if (prioritaire(instance, tachesRestantes.get(indexChallenger), tachesRestantes.get(indexBest))) {
                    indexBest = indexChallenger;
                }
            }
            return  this.tachesRestantes.remove(indexBest);
        }
    }

    @Override
    public ArrayList<Result> solve(Instance instance, long deadline) {
        //Initialisation
        jobFreeDate = new int[instance.numJobs];
        machineFreeDate = new int[instance.numMachines];
        int i;
        for (i=0; i<instance.numJobs; i++) {
            this.addTask(instance, new Task(i,0));
            this.jobFreeDate[i] = 0;
        }
        for (i=0; i<instance.numMachines; i++) {
            this.machineFreeDate[i] = 0;
        }
        ResourceOrder resourceOrder = new ResourceOrder(instance);
        Task task;
        Task nextTask;

        //Itérations
        while (!this.tachesRestantes.isEmpty() && deadline - System.currentTimeMillis() > 0) {
            task = this.getTask(instance);
            int startTime = getEarliestDate(instance, task);
            resourceOrder.addTaskToMachine(instance.machine(task), task);
            this.machineFreeDate[instance.machine(task)] = startTime + instance.duration(task);
            this.jobFreeDate[task.job] = startTime + instance.duration(task);
            nextTask = instance.nextTask(task);
            if (nextTask != null) {
                this.addTask(instance, new Task(task.job, task.task + 1));
            }
        }

        Result result;
        if (this.tachesRestantes.isEmpty()) {
            result = new Result(instance, resourceOrder.toSchedule(), Result.ExitCause.Blocked);
        } else {
            result = new Result(instance, resourceOrder.toSchedule(), Result.ExitCause.Timeout);
        }
        ArrayList<Result> list = new ArrayList<>();

        for (i=0; i<pts; i++) {
            list.add(result);
        }
        return list;
    }

    @Override
    public void setIterMax(int iterMax) {}

    @Override
    public void setpts(int pts) {this.pts = pts;}
}
